#include <math.h>
#include "common.h"

#ifdef TARGET_DOS

#include <allegro.h>

BEGIN_GFX_DRIVER_LIST
END_GFX_DRIVER_LIST
BEGIN_COLOR_DEPTH_LIST
END_COLOR_DEPTH_LIST
BEGIN_JOYSTICK_DRIVER_LIST
END_JOYSTICK_DRIVER_LIST

AUDIOSTREAM *outstr;

bool begin_sound_card(int freq, int bufsize, char *devicename)
{
	allegro_init();
	install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL);
	outstr = play_audio_stream(bufsize, 8, 0, freq, 255, 127);
	return outstr != NULL;
}

void end_sound_card(void)
{
	stop_audio_stream(outstr);
}

unsigned char *get_next_pointer_card(void)
{
	unsigned char *sndbuf;
	while(! (sndbuf = (unsigned char *)get_audio_stream_buffer(outstr)));
	return sndbuf;
}

void done_with_pointer_card(void)
{
	free_audio_stream_buffer(outstr);
}

#endif

#include <stdio.h>
#include <malloc.h>
FILE *outraw;
unsigned char *tmp;
int bufs;

bool begin_sound_file(int freq, int bufsize)
{
	printf("Writing a %dHz, 8bit, mono, unsigned RAW file named \"sound.raw\"\n", freq);
	outraw = fopen("sound.raw", "wb");
	bufs = bufsize;
	tmp = (unsigned char *)malloc(bufsize);
	return true;
}

void end_sound_file(void)
{
	fclose(outraw);
	free(tmp);
}

unsigned char *get_next_pointer_file(void)
{
	return tmp;
}

void done_with_pointer_file(void)
{
	fwrite(tmp, 1, bufs, outraw);
}

#ifdef TARGET_WIN
#include <windows.h>

HWAVEOUT device;
volatile WAVEHDR block1, block2, *cblock;
volatile unsigned char *data1, *data2;

bool begin_sound_card(int freq, int bufsize, char *devicename)
{
	WAVEFORMATEX format;

	format.wFormatTag = WAVE_FORMAT_PCM;
	format.nChannels = 1;
	format.nSamplesPerSec = freq;
	format.nAvgBytesPerSec = freq;
	format.nBlockAlign = 1;
	format.wBitsPerSample = 8;
	format.cbSize = 0;

	data1 = data2 = NULL;

	if(waveOutOpen(&device, WAVE_MAPPER, &format, 0, 0, CALLBACK_NULL) != MMSYSERR_NOERROR)
		return false;

	/* prepare header 1 */
	data1 = (volatile unsigned char *)malloc(bufsize);
	if(!data1)
		return false;

	block1.lpData = (char *)data1;
	block1.dwBufferLength = bufsize;
	block1.dwFlags = 0;

	if(waveOutPrepareHeader(device, (WAVEHDR *)&block1, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
	{
		free((char *)data1); data1 = NULL;
		return false;
	}

	/* prepare header 2 */
	data2 = (volatile unsigned char *)malloc(bufsize);
	if(!data2)
		return false;

	block2.lpData = (char *)data2;
	block2.dwBufferLength = bufsize;
	block2.dwFlags = 0;

	if(waveOutPrepareHeader(device, (WAVEHDR *)&block2, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
	{
		free((char *)data2); data2 = NULL;
		return false;
	}

	/* finish */
	memset((char *)data1, 127, bufsize);
	memset((char *)data2, 127, bufsize);

	waveOutWrite(device, (WAVEHDR *)&block2, sizeof(WAVEHDR));
	cblock = &block1;

	return true;
}

void end_sound_card(void)
{
	if(data1)
	{
		waveOutUnprepareHeader(device, (WAVEHDR *)&block1, sizeof(WAVEHDR));
		free((char *)data1);
	}

	if(data2)
	{
		waveOutUnprepareHeader(device, (WAVEHDR *)&block2, sizeof(WAVEHDR));
		free((char *)data2);
	}

	waveOutClose(device);
}

unsigned char *get_next_pointer_card(void)
{
	return (unsigned char *)cblock->lpData;
}

void done_with_pointer_card(void)
{
	volatile WAVEHDR *next;

	next = (cblock == &block1) ? &block2 : &block1;

	waveOutWrite(device, (WAVEHDR *)cblock, sizeof(WAVEHDR));
	while(!(next->dwFlags&WHDR_DONE));

	cblock = next;
}

#endif

#ifdef TARGET_OSS_OLD

#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>

int audio_fd;
int bufsize;
unsigned char *audio_buffer;

bool begin_sound_card(int freq, int b, char *devicename)
{
	int speed = 44100;
	int *sigs;

	bufsize = b;

	if((audio_fd = open("/dev/dsp1", O_WRONLY, 0)) == -1)
		return false;

	if(ioctl(audio_fd, SNDCTL_DSP_SPEED, &speed) == -1)
	{
		close(audio_fd);
		return false;
	}

	if(abs(speed-44100) > 2205)
	{
		close(audio_fd);
		return false;
	}

	audio_buffer = (unsigned char *)calloc(bufsize, 1);

	return true;
}

void end_sound_card(void)
{
	free(audio_buffer);
	close(audio_fd);
}

unsigned char *get_next_pointer_card(void)
{
	return audio_buffer;
}

void done_with_pointer_card(void)
{
	while(!write(audio_fd, audio_buffer, bufsize));
}

#endif
#ifdef TARGET_ALSA

#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
//#include <sys/soundcard.h>
#include <stdio.h>
#include <alsa/asoundlib.h>
#include <inttypes.h>

snd_pcm_t *playback_handle;
snd_pcm_hw_params_t *hw_params;
snd_pcm_uframes_t frames;
int audio_fd;
int bufsize;
int buff_size;
int hardware_pausable;
//unsigned char *audio_buffer;
short *audio_buffer;
int err;
unsigned int tmp2;
int seconds;

bool begin_sound_card(int freq, int b, char *devicename)
{
	unsigned int speed = 44100;
	unsigned int val;
	int *sigs;

	bufsize = b;
	if (DEBUG) printf("begin_sound_card: bufsize set to %d\n",b);

	if (devicename == NULL) {
		err = snd_pcm_open (&playback_handle, "plughw:1,0", SND_PCM_STREAM_PLAYBACK, 0);
	} else {
		err = snd_pcm_open (&playback_handle, devicename, SND_PCM_STREAM_PLAYBACK, 0);
	}

	if (err < 0) {
		fprintf (stderr, "cannot open audio device %s (%s)\n",
		 snd_strerror (err));
		exit (1);
	}
		   
	if ((err = snd_pcm_hw_params_malloc (&hw_params)) < 0) {
		fprintf (stderr, "cannot allocate hardware parameter structure (%s)\n",
		 snd_strerror (err));
		exit (1);
	}
				 
	if ((err = snd_pcm_hw_params_any (playback_handle, hw_params)) < 0) {
		fprintf (stderr, "cannot initialize hardware parameter structure (%s)\n",
			 snd_strerror (err));
		exit (1);
	}
	
	if ((err = snd_pcm_hw_params_set_access (playback_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
		fprintf (stderr, "cannot set access type (%s)\n",
			 snd_strerror (err));
		exit (1);
	}
	
	if ((err = snd_pcm_hw_params_set_format (playback_handle, hw_params, SND_PCM_FORMAT_S16_LE)) < 0) {
		fprintf (stderr, "cannot set sample format (%s)\n",
			 snd_strerror (err));
		exit (1);
	}
	

	int subunitdirection = 0;
	if ((err = snd_pcm_hw_params_set_rate_near (playback_handle, hw_params, &speed, &subunitdirection)) < 0) {
		fprintf (stderr, "cannot set sample rate (%s)\n",
			 snd_strerror (err));
		exit (1);
	}
	
	if ((err = snd_pcm_hw_params_set_channels (playback_handle, hw_params, 2)) < 0) {
		fprintf (stderr, "cannot set channel count (%s)\n",
		 snd_strerror (err));
		exit (1);
	}
	
	if ((err = snd_pcm_hw_params (playback_handle, hw_params)) < 0) {
		fprintf (stderr, "cannot set parameters (%s)\n",
			 snd_strerror (err));
		exit (1);
	}
	
	
	if ((err = snd_pcm_prepare (playback_handle)) < 0) {
		fprintf (stderr, "cannot prepare audio interface for use (%s)\n",
			 snd_strerror (err));
		exit (1);
	}

	if (DEBUG) printf("PCM name: '%s'\n", snd_pcm_name(playback_handle));
 
	if (DEBUG) printf("PCM state: %s\n", snd_pcm_state_name(snd_pcm_state(playback_handle)));
 
	snd_pcm_hw_params_get_channels(hw_params, &tmp2);
	if (DEBUG) printf("channels: %i ", tmp2);
 
	if (tmp2 == 1)
		if (DEBUG) printf("(mono)\n");
	else if (tmp2 == 2)
		if (DEBUG) printf("(stereo)\n");
 
	snd_pcm_hw_params_get_rate(hw_params, &tmp2, 0);
	if (DEBUG) printf("rate: %d bps\n", tmp2);
 
	/* Allocate buffer to hold single period */
	snd_pcm_hw_params_get_period_size(hw_params, &frames, 0);
 
	// divide frames by 4
	frames = frames >>2;
	if (DEBUG) printf("frames = %d\n",frames);
	int channels = 2;
	buff_size = frames * channels * 2;
	if (DEBUG) printf("buff_size = %d\n", buff_size);
	//audio_buffer = (unsigned char *) malloc(buff_size);
	audio_buffer = (short *) malloc(buff_size);
 
	snd_pcm_hw_params_get_period_time(hw_params, &tmp2, NULL);
 
	hardware_pausable=0;
	val = snd_pcm_hw_params_can_pause(hw_params);
	if (val) {
		hardware_pausable=1;
	}
  	if (DEBUG) printf("can pause = %d\n", val);
	
	snd_pcm_hw_params_free (hw_params);
	//audio_buffer = (unsigned char *)calloc(bufsize, 1);

	return true;
}

void end_sound_card(void)
{
	printf("Waiting for sound buffer to flush\n");
	// doing a drop will actually abort the audio too early
  	//printf("snd_pcm_drop(playback_handle)\n");
  	//snd_pcm_drop(playback_handle);
  	if (DEBUG) printf("snd_pcm_drain(playback_handle)\n");
  	snd_pcm_drain(playback_handle);
  	if (DEBUG) printf("snd_pcm_close(playback_handle)\n");
  	snd_pcm_close(playback_handle);
	free(audio_buffer);
	//close(audio_fd);
}

//unsigned char *get_next_pointer_card(void)
short *get_next_pointer_card(void)
{
	return audio_buffer;
}

unsigned int get_buff_size(void) {
	return buff_size;
}

void done_with_pointer_card(void)
{

	err = snd_pcm_writei (playback_handle, audio_buffer, frames);
	if (err>0)
	{
		//fprintf (stderr, "snd_pcm_writei: wrote only %d\n",err); 
		fflush(stderr);
	}
	else
	{
		if (err == -32)								// if broken pipe -32
		{
               		//soundoutput_underruns++;
			int errb;
			errb=snd_pcm_recover(playback_handle,err,0);			// error passed is resullt of writen
			if (errb<0)
			{
				fprintf(stderr,"Failed to recover from underrun\n");
				fflush(stderr);
				exit(1);
			}
			else {
				printf("RECOVER!");
			}
		}
	}
	//while(!write(audio_fd, audio_buffer, bufsize));
}

void pause_sound_card(void) {
	//snd_pcm_drop (playback_handle);
	if (hardware_pausable) {
		snd_pcm_pause(playback_handle,1);
	} else {
		printf("Your sound hardware does not support pausing\n");
	}

}

void resume_sound_card(void) {
	if (hardware_pausable) {
		snd_pcm_pause(playback_handle,0);
	}
}

#endif
